%Finn Haugen (finn@techteach.no)
%----------------------------------------------------------
disp('-------------------')
disp('Using fmincon() in Matlab solve an optimization')
disp('(Nonlinear Programming, NLP) problem:')
disp('min_x f(x)')
disp('s.t.')
disp('model: f= AC_heater_error(x1,x2,x3,x4,data)')
disp('constraints:')
disp('1 <= x1 <= 10')
disp('10 <= x2 <= 100')
disp('1 <= x3 <= 20')
disp('15 <= x4 <= 30')
%----------------------------------------------------------
format compact
clear all
close all
%--------------------------------------------------
% Load the data 
global u_estim y_estim
data_sys = load('airheater_logfile.txt');

%% Selecting the estimation and validation dataset
L=length(data_sys);
F=0.9;%First fraction (portion) of logfile to be used for model adaption:
N=floor(L*F); %N is number of samples used for estimation.
del_t=0.1;

%Rest used for visual validation.
t_estim=del_t*[1:N]';
u_estim=data_sys(1:N,2);
y_estim=data_sys(1:N,3);
t_valid=del_t*[N+1:L]';
u_valid=data_sys(N+1:L,2);
y_valid=data_sys(N+1:L,3);
t_total=del_t*[1:L]';
u_total=data_sys(1:L,2);
y_total=data_sys(1:L,3);

%% Parameter estimation using Nonlinear least squares (fmincon)
%Contraints:
x_lb=[1,10,1,15]';
x_ub=[10,100,20,30]';
%--------------------------------------------------
%fmincon:
%x_guess=[4,20,2,20]';

x_guess=[3,23,3,21]';

Aineq=[]; Bineq=[]; Aeq=[]; Beq=[];
tic
fun_objective_handle = @(x)fun_objective(x);%Local function
fun_constraints_handle = @(x)fun_constraints(x);%Local function
optim_options=optimset('Display','on');
% optim_options=optimset('Algorithm','sqp','Display','on');
[x_opt,fval,exitflag,output,lambda,grad,hessian] =...
fmincon(fun_objective_handle,x_guess,Aineq,Bineq,Aeq,Beq,x_lb,x_ub,...
    fun_constraints_handle,optim_options);
toc
%--------------------------------------------------

%Displaying the optimal solution:
disp('-------------------')
disp('Optimal solution:')

fval
x_opt
K_opt=x_opt(1)
thetaT_opt=x_opt(2)
thetaD_opt=x_opt(3)
T_env=x_opt(4)

%% Model Validation
del_t=0.1;
A=(1-del_t/thetaT_opt);
B=[del_t*K_opt/thetaT_opt del_t/thetaT_opt] ;
C=1;
D=[0 0];
model_est=ss(A,B,C,D,del_t,'InputDelay',round(thetaD_opt/del_t));
u_total=[u_total T_env.*ones(length(u_total),1)];
x0=y_total(1);
[y_sim,t,x] = lsim(model_est,u_total,[],x0);

%% Plots
h=figure; %Getting figure handle
fig_posleft=8;fig_posbottom=1.5;fig_width=24;fig_height=20;
fig_pos_size_1=[fig_posleft,fig_posbottom,fig_width,fig_height];

set(gcf,'Units','centimeters','Position',fig_pos_size_1);
figtext='Estimation of model params of Air-Heater';
set(gcf,'Name',figtext,'NumberTitle','on')

subplot(2,1,1)
figure(1)
plot(t_total,y_sim,'r',t_total,y_total,'b');
title('Real y (blue). Simulated y with adapted model (red). Green: Interval for adaption.')
grid minor
ylim([15,35]);
ylabel('[Temperature]');xlabel('t [s]')

subplot(2,1,2)
plot(t_total,u_total,'b');
title('Applied control sigal, u, to both real and simulated process. Green: Interval for adaptation.')
grid minor
ylim([0,4]);
ylabel('[V]');xlabel('t [s]')

%% Function for calculating the error
function f = fun_objective(x)
global u_estim y_estim
Kh=x(1);
theta_t=x(2);
dt=0.1;
theta_d=round(x(3)/dt)+1;

Tprev_meas=y_estim(1:end-1);
Tout_meas=y_estim(2:end);
u=u_estim(1:end);
u = [u(1)*ones(theta_d,1); u(1:end-theta_d-1)];
T_env=x(4).*ones(length(u),1);
T_est=((1-dt/theta_t).*Tprev_meas)+((Kh*dt/theta_t).*u)+(dt/theta_t.*T_env);
error=Tout_meas-T_est;
f=error'*error
end

% Defining constraints
function [cineq,ceq]=fun_constraints(x)
cineq = [];  % Compute nonlinear inequalities.
% cineq =  -x(2)+x(1)+2;   % Compute nonlinear inequalities.
    %Inequality constraint: x2 >= x1+2
ceq = [];   % Compute nonlinear equalities.
end